/* 
 * windraw.c --
 *
 *	This module implements the "window" and "drawable"
 *	subcommands of the xop command.
 */

/*
 * Copyright (C) 1998-2001, Neil McKay
 */

#include "tk.h"

/* Include all the X11 stuff we need. */

#include <X11/Xutil.h>

#include "xopInt.h"



/*
 *--------------------------------------------------------------
 *
 * ConvertXColorToRGBBytes --
 *
 *	This procedure converts the red, green and blue values
 *	stored in an XColor structure to unsigned 8-bit values
 *	and stores them in an array of unsigned bytes. We have
 *	bytesPtr[0] = red, bytesPtr[1] = green, and
 *	bytesPtr[2] = blue.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static void
ConvertXColorToRGBBytes(xColorPtr, bytesPtr)
    XColor *xColorPtr;		/* XColor that we'll convert		*/
    unsigned char *bytesPtr;	/* Bytes where we'll put the color info	*/
{

    if (xColorPtr->red > 0xff00) {
	bytesPtr[0] = 0xff;
    } else {
	bytesPtr[0] = (xColorPtr->red + 0x80) >> 8;
    }
    if (xColorPtr->green > 0xff00) {
	bytesPtr[1] = 0xff;
    } else {
	bytesPtr[1] = (xColorPtr->green + 0x80) >> 8;
    }
    if (xColorPtr->blue > 0xff00) {
	bytesPtr[2] = 0xff;
    } else {
	bytesPtr[2] = (xColorPtr->blue + 0x80) >> 8;
    }
}


/*
 *--------------------------------------------------------------
 *
 * PutDrawableIntoPhoto --
 *
 *	Copies the contents of a drawable into a Tk photo image.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A Tk photo image gets modified.
 *
 *--------------------------------------------------------------
 */

#define	MAX_CMAP_LOAD_DEPTH	12	/* Largest image depth where
					 * we'll allow loading of
					 * the entire colormap
					 */
static int
PutDrawableIntoPhoto(interp, display, screenNo, colormap, drawableID, imageName)
    Tcl_Interp *interp;		/* Error messages go here	*/
    Display *display;		/* Display on which drawable
				 * is found			*/
    int screenNo;		/* Number of screen on which
				 * to find default colormap
				 * (Only needed if colormap
				 * is "None")			*/
    Colormap colormap;		/* Colormap used to interpret
				 * drawable's colors		*/
    Drawable drawableID;	/* Drawable to copy		*/
    char *imageName;		/* Name of image to create	*/
{
    XopDrawableGeometry geom;
    Tk_ErrorHandler handler;
    int gotXError;
    int result;
    XImage *xImagePtr;
    XColor color;
    Tk_PhotoHandle photoImage;
    Tk_PhotoImageBlock photoBlock;
    int nPixelsInImage;
    unsigned char *photoPixPtr;
    int nCmapEntries;
    XColor *colorArray;
    unsigned long pixelVal;
    int row, col, i;
    char junkStr[64];

    /* Make sure that the photo image exists. */

    photoImage = Tk_FindPhoto(interp, imageName);
    if (photoImage == (Tk_PhotoHandle) NULL) {
	Tcl_AppendResult(interp, "No photo image named \"",
			imageName, "\"", (char *) NULL);
	return TCL_ERROR;
    }

    /* Get the characteristics of the drawable. */

    result = XopGetDrawableGeometry(display, drawableID, &geom);
    if (result == 0) {
	Tcl_SetResult(interp,
		"couldn't get drawable's geometry", TCL_STATIC);
	return TCL_ERROR;
    }

    /* Try to get the image. */

    gotXError = 0;
    handler = Tk_CreateErrorHandler(display, -1, -1, -1,
		XopXErrorHandler, (ClientData) (&gotXError));
    xImagePtr = XGetImage(display, drawableID,
			0, 0, geom.width, geom.height, AllPlanes, ZPixmap);
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);

    /* Did we get succeed? */

    if (xImagePtr == (XImage *) NULL || gotXError) {
	Tcl_SetResult(interp, "XGetImage failed", TCL_STATIC);
	return TCL_ERROR;
    }

    /* "Empty out" the photo image. */

    Tk_PhotoBlank(photoImage);
#if TK_MINOR_VERSION >= 5
    result = Tk_PhotoSetSize(interp, photoImage,
				xImagePtr->width, xImagePtr->height);
    if(result != TCL_OK) {
	return result;
    }
#else
    Tk_PhotoSetSize(photoImage, xImagePtr->width, xImagePtr->height);
#endif

    /*
     * If the depth is greater than 1 and the colormap
     * is "None", use the screen's default colormap.
     */

    if (colormap == None && geom.depth >= 1) {
	colormap = XDefaultColormap(display, screenNo);
    }

    /*
     * Allocate a temporary data array; we'll only do
     * one PhotoPutBlock call.
     */

    nPixelsInImage = (xImagePtr->width)*(xImagePtr->height);
    photoBlock.pixelPtr = (unsigned char *) ckalloc(nPixelsInImage*3);
    if (photoBlock.pixelPtr == (unsigned char *) NULL) {
	XDestroyImage(xImagePtr);
	Tcl_SetResult(interp,
		"Not enough memory to create image", TCL_STATIC);
	return TCL_ERROR;
    }

    photoBlock.width = xImagePtr->width;
    photoBlock.height = xImagePtr->height;
    photoBlock.pitch = 3*(xImagePtr->width);
    photoBlock.pixelSize = 3;
    photoBlock.offset[0] = 0;
    photoBlock.offset[1] = 1;
    photoBlock.offset[2] = 2;

    /*
     * At this point, we should either have a valid colormap,
     * or a colormap of "None" and a single-plane image.
     * If the colormap is given (i.e. it's not "None"), use it.
     */

    if (colormap == None) {
	photoPixPtr = photoBlock.pixelPtr;
	for(row = 0; row < xImagePtr->width; row++) {
	    for(col = 0; col < xImagePtr->width; col++) {
		color.pixel = XGetPixel(xImagePtr, col, row);
		if(color.pixel == 0) {
		    photoPixPtr[0] = photoPixPtr[1] = photoPixPtr[2] = 0x00;
		} else {
		    photoPixPtr[0] = photoPixPtr[1] = photoPixPtr[2] = 0xff;
		}
		photoPixPtr += 3;
	    }
	}
    } else if (geom.depth <= MAX_CMAP_LOAD_DEPTH) {
	/*
	 * At this point, we know we have a valid colormap. If
	 * the image depth is small enough, load in the whole
	 * colormap; this way, we don't have to do a round-trip
	 * to the server for each pixel.
	 */

	nCmapEntries = (1 << geom.depth);
	colorArray = (XColor *) ckalloc(nCmapEntries*sizeof(XColor));
	if (colorArray == (XColor *) NULL) {
	    ckfree((VOID *) (photoBlock.pixelPtr));
	    XDestroyImage(xImagePtr);
	    Tcl_SetResult(interp,
			"Not enough memory to map image colors", TCL_STATIC);
	    return TCL_ERROR;
	}

	for(i = 0; i < nCmapEntries; i++) {
	    colorArray[i].pixel = i;
	    colorArray[i].flags = (DoRed|DoGreen|DoBlue);
	}

	if (XopQueryColors(display, colormap,
			colorArray, nCmapEntries) != 0) {
	    ckfree((VOID *) colorArray);
	    ckfree((VOID *) (photoBlock.pixelPtr));
	    XDestroyImage(xImagePtr);
	    Tcl_SetResult(interp,
			"Could not read colormap", TCL_STATIC);
	    return TCL_ERROR;
	}

	/* Get each pixel... */

	photoPixPtr = photoBlock.pixelPtr;
	for(row = 0; row < xImagePtr->height; row++) {
	    for(col = 0; col < xImagePtr->width; col++) {
		pixelVal = XGetPixel(xImagePtr, col, row);
		ConvertXColorToRGBBytes(&(colorArray[pixelVal]), photoPixPtr);
		photoPixPtr += 3;
	    }
	}
	ckfree((VOID *) colorArray);
    } else {
	/*
	 * @@@ For now, we'll do things the slow, but simple and reliable,
	 * @@@ way. Later we should change this so that it doesn't have
	 * @@@ to do a round trip to the server for each pixel.
	 */

	photoPixPtr = photoBlock.pixelPtr;
	for(row = 0; row < xImagePtr->height; row++) {
	    for(col = 0; col < xImagePtr->width; col++) {
		color.pixel = XGetPixel(xImagePtr, col, row);
		if (XopQueryColor(display, colormap, &color) != TCL_OK) {
		    ckfree((VOID *) (photoBlock.pixelPtr));
		    XDestroyImage(xImagePtr);
		    Tcl_ResetResult(interp);
		    sprintf(junkStr, "%1ld (0x%lx)", color.pixel, color.pixel);
		    Tcl_AppendResult(interp,
			"Couldn't translate color for pixel value ",
			junkStr, (char *) 0);
		    return TCL_ERROR;
		}
		ConvertXColorToRGBBytes(&color, photoPixPtr);
		photoPixPtr += 3;
	    }
	}
    }

    /*
     * We've got the array of pixels: put 'em in
     * the photo image, clean up, and return.
     */

#if TK_MINOR_VERSION >= 5
    result = Tk_PhotoPutBlock(interp, photoImage, &photoBlock,
			0, 0, xImagePtr->width, xImagePtr->height,
			TK_PHOTO_COMPOSITE_SET);
#else
#if TK_MINOR_VERSION == 4
    Tk_PhotoPutBlock(photoImage, &photoBlock,
			0, 0, xImagePtr->width, xImagePtr->height,
			TK_PHOTO_COMPOSITE_SET);
#else
    Tk_PhotoPutBlock(photoImage, &photoBlock,
			0, 0, xImagePtr->width, xImagePtr->height);
#endif
    result = TCL_OK;
#endif

    ckfree((VOID *) photoBlock.pixelPtr);
    XDestroyImage(xImagePtr);
    return result;
}


/*
 *--------------------------------------------------------------
 *
 * XopDrawableCommand --
 *
 *	Implements the "drawable" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopDrawableCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    Drawable drawableId;
    int result;
    char junkStr[32];
    XopDrawableGeometry geom;

    if (argc < 4) {
	Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" <drawableID> geometry|getimage|killclient ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    /* Get the drawable ID. */

    if (XopGetDrawableId(interp, argv[2], &drawableId) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Do the commands... */

    if (strcmp(argv[3], "geometry") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp,
		    "Usage: ", argv[0], " ", argv[1],
		    " <drawableID> geometry",
		     (char *) NULL);
	    return TCL_ERROR;
	}

	result = XopGetDrawableGeometry(opts->display, drawableId, &geom);
	if (result == 0) {
	    Tcl_SetResult(interp,
			"Couldn't get drawable geometry", TCL_STATIC);
	    return TCL_ERROR;
	}

	Tcl_AppendElement(interp, "x");
	sprintf(junkStr, "%1d", geom.x);
	Tcl_AppendElement(interp, junkStr);

	Tcl_AppendElement(interp, "y");
	sprintf(junkStr, "%1d", geom.y);
	Tcl_AppendElement(interp, junkStr);

	Tcl_AppendElement(interp, "width");
	sprintf(junkStr, "%1d", geom.width);
	Tcl_AppendElement(interp, junkStr);

	Tcl_AppendElement(interp, "height");
	sprintf(junkStr, "%1d", geom.height);
	Tcl_AppendElement(interp, junkStr);

	Tcl_AppendElement(interp, "borderwidth");
	sprintf(junkStr, "%1d", geom.border_width);
	Tcl_AppendElement(interp, junkStr);

	Tcl_AppendElement(interp, "depth");
	sprintf(junkStr, "%1d", geom.depth);
	Tcl_AppendElement(interp, junkStr);

	return TCL_OK;
    }

    if (strcmp(argv[3], "getimage") == 0) {
	if (argc != 5) {
	    Tcl_AppendResult(interp,
		    "Usage: ", argv[0], " ", argv[1],
		    " <drawableID> getimage <imageName>",
		     (char *) NULL);
	    return TCL_ERROR;
	}

	return(PutDrawableIntoPhoto(interp, opts->display, opts->screenNumber,
		opts->colormap, drawableId, argv[4]));
    }

    if (strcmp(argv[3], "killclient") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp,
		    "Usage: ", argv[0], " ", argv[1],
		    " <drawableID> killclient",
		     (char *) NULL);
	    return TCL_ERROR;
	}

	/*@@@ Should trap errors here. @@@*/
	XKillClient(opts->display, drawableId);
	return TCL_OK;
    }

    Tcl_AppendResult(interp,
		"Usage: ", argv[0], " ", argv[1],
		" <drawableID> geometry|getimage|killclient ...",
		(char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * ListChildren--
 *
 *	This procedure appends a list of children of
 *	window "Xwin" of display "display". The child
 *	window IDs are written as hex numbers.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static int
ListChildren(interp,display,Xwin)
    Tcl_Interp *interp;		/* Result appended here		*/
    Display *display;		/* Display where window resides	*/
    Window Xwin;		/* Window to find children of	*/
{
    Window rootRet;
    Window parentRet;
    Window *childList;
    int i;
    unsigned int nChildren;
    Tk_ErrorHandler handler;
    int gotXError;
    Status status;

    /* Try to get a list of child window IDs. */

    childList = (Window *) NULL;
    gotXError = 0;
    handler = Tk_CreateErrorHandler(display, -1, -1, -1,
		XopXErrorHandler, (ClientData) (&gotXError));
    status = XQueryTree(display, Xwin, &rootRet,
	    &parentRet, &childList, &nChildren);
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);

    if (status == 0 || gotXError) {
	if(childList != (Window *) NULL) {
	    XFree(childList);
	}

	Tcl_SetResult(interp, "XQueryTree failed", TCL_STATIC);
	return(TCL_ERROR);
    }

    /* Generate a properly-formatted list of X window IDs. */

    for(i = 0; i < nChildren; i++) {
	XopAppendDrawableId(interp,(Drawable) (childList[i]));
    }

    /* Clean up and return. */

    XFree(childList);
    return(TCL_OK);
}


/*
 *--------------------------------------------------------------
 *
 * AppendParentXId --
 *
 *	This procedure appends the parent of window
 *	"Xwin" of display "display" onto "interp"'s
 *	result string. The parent window ID is written
 *	as a hex number.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static int
AppendParentXId(interp,display,Xwin)
    Tcl_Interp *interp;		/* Result appended here		*/
    Display *display;		/* Display where window resides	*/
    Window Xwin;		/* Window to find parent of	*/
{
    Window parent;

    /* Try to get the parent window ID. */

    parent = XopParentXId(display, Xwin);
    if (parent == None) {
	Tcl_SetResult(interp, "Couldn't find parent", TCL_STATIC);
	return(TCL_ERROR);
    }

    /*  Append the parent's window ID. */

    XopAppendDrawableId(interp, (Drawable) parent);
    return(TCL_OK);
}



/*
 *--------------------------------------------------------------
 *
 * WriteWinAttr --
 *
 *	Write a window's attributes into a Tcl interpreter's
 *	result string, in text format.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	interp's result string is modified.
 *
 *--------------------------------------------------------------
 */

static void
WriteWinAttr(interp, attrPtr)
    Tcl_Interp *interp;	/* Text version of attributes goes here	*/
    XWindowAttributes *attrPtr;	/* attributes to convert	*/
{
    char junkStr[32];

    /* Position */

    Tcl_AppendElement(interp, "position");
    sprintf(junkStr, "%1d %1d", attrPtr->x, attrPtr->y);
    Tcl_AppendElement(interp, junkStr);

    /* Size */

    Tcl_AppendElement(interp, "size");
    sprintf(junkStr, "%1d %1d", attrPtr->width, attrPtr->height);
    Tcl_AppendElement(interp, junkStr);

    /* Border width */

    Tcl_AppendElement(interp, "borderWidth");
    sprintf(junkStr, "%1d", attrPtr->border_width);
    Tcl_AppendElement(interp, junkStr);

    /* Depth */

    Tcl_AppendElement(interp, "depth");
    sprintf(junkStr, "%1d", attrPtr->depth);
    Tcl_AppendElement(interp, junkStr);

    /* Root window of screen containing window */

    Tcl_AppendElement(interp, "root");
    XopAppendDrawableId(interp, attrPtr->root);

    /* Class */

    Tcl_AppendElement(interp, "class");
    if (attrPtr->class == InputOutput) {
	Tcl_AppendElement(interp, "inputoutput");
    } else if (attrPtr->class == InputOnly) {
	Tcl_AppendElement(interp, "inputonly");
    } else {
	sprintf(junkStr, "%1d", attrPtr->class);
	Tcl_AppendElement(interp, junkStr);
    }

    /* Bit gravity */

    Tcl_AppendElement(interp, "bitGravity");
    if(attrPtr->bit_gravity == ForgetGravity) {
        Tcl_AppendElement(interp, "forget");
    } else if(attrPtr->bit_gravity == StaticGravity) {
        Tcl_AppendElement(interp, "static");
    } else if(attrPtr->bit_gravity == CenterGravity) {
        Tcl_AppendElement(interp, "c");
    } else if(attrPtr->bit_gravity == EastGravity) {
        Tcl_AppendElement(interp, "e");
    } else if(attrPtr->bit_gravity == NorthEastGravity) {
        Tcl_AppendElement(interp, "ne");
    } else if(attrPtr->bit_gravity == NorthGravity) {
        Tcl_AppendElement(interp, "n");
    } else if(attrPtr->bit_gravity == NorthWestGravity) {
        Tcl_AppendElement(interp, "nw");
    } else if(attrPtr->bit_gravity == SouthEastGravity) {
        Tcl_AppendElement(interp, "se");
    } else if(attrPtr->bit_gravity == SouthGravity) {
        Tcl_AppendElement(interp, "s");
    } else if(attrPtr->bit_gravity == SouthWestGravity) {
        Tcl_AppendElement(interp, "sw");
    } else if(attrPtr->bit_gravity == WestGravity) {
        Tcl_AppendElement(interp, "w");
    } else {
	sprintf(junkStr, "%1d", attrPtr->bit_gravity);
	Tcl_AppendElement(interp, junkStr);
    }

    /* Window gravity */

    Tcl_AppendElement(interp, "winGravity");
    if(attrPtr->win_gravity == UnmapGravity) {
        Tcl_AppendElement(interp, "unmap");
    } else if(attrPtr->win_gravity == StaticGravity) {
        Tcl_AppendElement(interp, "static");
    } else if(attrPtr->win_gravity == CenterGravity) {
        Tcl_AppendElement(interp, "c");
    } else if(attrPtr->win_gravity == EastGravity) {
        Tcl_AppendElement(interp, "e");
    } else if(attrPtr->win_gravity == NorthEastGravity) {
        Tcl_AppendElement(interp, "ne");
    } else if(attrPtr->win_gravity == NorthGravity) {
        Tcl_AppendElement(interp, "n");
    } else if(attrPtr->win_gravity == NorthWestGravity) {
        Tcl_AppendElement(interp, "nw");
    } else if(attrPtr->win_gravity == SouthEastGravity) {
        Tcl_AppendElement(interp, "se");
    } else if(attrPtr->win_gravity == SouthGravity) {
        Tcl_AppendElement(interp, "s");
    } else if(attrPtr->win_gravity == SouthWestGravity) {
        Tcl_AppendElement(interp, "sw");
    } else if(attrPtr->win_gravity == WestGravity) {
        Tcl_AppendElement(interp, "w");
    } else {
	sprintf(junkStr, "%1d", attrPtr->win_gravity);
	Tcl_AppendElement(interp, junkStr);
    }

    /* Colormap */

    Tcl_AppendElement(interp, "colormap");
    sprintf(junkStr, "0x%x", (unsigned int) (attrPtr->colormap));
    Tcl_AppendElement(interp, junkStr);

    /* Colormap installed? */

    Tcl_AppendElement(interp, "mapInstalled");
    if(attrPtr->map_installed) {
        Tcl_AppendElement(interp, "1");
    } else {
        Tcl_AppendElement(interp, "0");
    }

    /* Override-redirect */

    Tcl_AppendElement(interp, "overrideRedirect");
    sprintf(junkStr, "%1d", attrPtr->override_redirect);
    Tcl_AppendElement(interp, junkStr);

    /* Mapping State */

    Tcl_AppendElement(interp, "mapState");
    if(attrPtr->map_state == IsUnmapped) {
        Tcl_AppendElement(interp, "unmapped");
    } else if(attrPtr->map_state == IsUnviewable) {
        Tcl_AppendElement(interp, "unviewable");
    } else if(attrPtr->map_state == IsViewable) {
        Tcl_AppendElement(interp, "viewable");
    } else {
	sprintf(junkStr, "%1d", attrPtr->map_state);
	Tcl_AppendElement(interp, junkStr);
    }
}


/*
 *--------------------------------------------------------------
 *
 * XopWindowCommand --
 *
 *	Implements the "window" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int XopWindowCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    int result;
    Window windowID;
    XWindowAttributes winAttr;
    Tk_Window tkWinPtr;
    int x, y, width, height;

    if (argc < 4) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" windowID subcommand ...", (char *) NULL);
	return TCL_ERROR;
    }

    /* Get the window ID. */

    result = XopGetAndCheckWindowId(interp,opts->display,argv[2],&windowID);
    if (result != TCL_OK) {
	return TCL_ERROR;
    }

    /* Execute subcommands... */

    if (strcmp(argv[3], "attributes") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	result = XopGetWindowAttributes(opts->display,
			windowID, &winAttr);
	if (result == 0) {
	    Tcl_AppendResult(interp, "No window ",
			argv[2], " on display", (char *) 0);
	    return TCL_ERROR;
	}

	WriteWinAttr(interp, &winAttr);
	return TCL_OK;
    }

    if (strcmp(argv[3], "children") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	result = ListChildren(interp, opts->display, windowID);
	if (result != TCL_OK) {
	    result = TCL_ERROR;
	}

	return result;
    }

    if (strcmp(argv[3], "circulate") == 0) {
	if (argc != 6 || (strcmp(argv[4], "subwindows") != 0)) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
			argv[1], " windowID ", argv[3],
			" subwindows up|down", (char *) NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[5], "up") == 0) {
	    XCirculateSubwindowsUp(opts->display, windowID);
	    return TCL_OK;
	}
	if (strcmp(argv[5], "down") == 0) {
	    XCirculateSubwindowsDown(opts->display, windowID);
	    return TCL_OK;
	}

	Tcl_AppendResult(interp,
		"direction must be \"up\" or \"down\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[3], "configure") == 0) {
	/*@@@ N.Y.I. @@@*/

	Tcl_AppendResult(interp, argv[3],
		"subcommand not implemented", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[3], "destroy") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	XDestroyWindow(opts->display, windowID);
	return TCL_OK;
    }

    if (strcmp(argv[3], "destroysubwindows") == 0) {
	/*@@@ N.Y.I. @@@*/

	Tcl_AppendResult(interp, argv[3],
		"subcommand not implemented", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[3], "inthisapp") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	tkWinPtr = Tk_IdToWindow(opts->display, windowID);
	if(tkWinPtr == NULL) {
	    Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
	    Tcl_SetResult(interp, "1", TCL_STATIC);
	}

	return TCL_OK;
    }

    if (strcmp(argv[3], "lower") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	XLowerWindow(opts->display, windowID);
	return TCL_OK;
    }

    if (strcmp(argv[3], "map") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	XMapWindow(opts->display, windowID);
	return TCL_OK;
    }

    if (strcmp(argv[3], "map&raise") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	XMapRaised(opts->display, windowID);
	return TCL_OK;
    }

    if (strcmp(argv[3], "mapsubwindows") == 0) {
	/*@@@ N.Y.I. @@@*/

	Tcl_AppendResult(interp, argv[3],
		"subcommand not implemented", (char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[3], "move") == 0) {
	if (argc != 6) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0],
			" ", argv[1], " windowId ", argv[3],
			" move x y", (char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[4], &x) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[5], &y) != TCL_OK) {
	    return TCL_ERROR;
	}

	XMoveWindow(opts->display, windowID, x, y);
	return TCL_OK;
    }

    if (strcmp(argv[3], "move&resize") == 0) {
	if (argc != 8) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0],
			" ", argv[1], " windowId ", argv[3],
			" move&resize x y width height", (char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[4], &x) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[5], &y) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[6], &width) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[7], &height) != TCL_OK) {
	    return TCL_ERROR;
	}

	XMoveResizeWindow(opts->display, windowID, x, y, width, height);
	return TCL_OK;
    }

    if (strcmp(argv[3], "parent") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	result = AppendParentXId(interp, opts->display, windowID);
	if (result != TCL_OK) {
	    result = TCL_ERROR;
	}

	return result;
    }

    if (strcmp(argv[3], "property") == 0) {
	/*@@@ N.Y.I. @@@*/
    }

    if (strcmp(argv[3], "resize") == 0) {
	if (argc != 6) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0],
			" ", argv[1], " windowId ", argv[3],
			" resize width height", (char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[4], &width) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[5], &height) != TCL_OK) {
	    return TCL_ERROR;
	}

	XResizeWindow(opts->display, windowID, width, height);
	return TCL_OK;
    }

    if (strcmp(argv[3], "unmap") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "Extraneous arguments follow \"",
						argv[3], "\"", (char *) NULL);
	    return TCL_ERROR;
	}

	XUnmapWindow(opts->display, windowID);
	return TCL_OK;
    }

    Tcl_AppendResult(interp, "Unknown subcommand \"", argv[3],
		"\"; legal subcommands are:\n",
		"\tattributes children circulate configure\n",
		"\tdestroy destroysubwindows inthisapp lower\n",
		"\tmap map&raise mapsubwindows move move&resize\n",
		"\tparent property raise resize unmap",
		(char *) NULL);
    return TCL_ERROR;
}

